package com.burtbeckwith.gae.framework

import org.apache.commons.logging.Log
import org.apache.commons.logging.LogFactory

import javax.jdo.JDOHelper
import javax.jdo.JDOObjectNotFoundException
import javax.jdo.PersistenceManager
import javax.jdo.PersistenceManagerFactory
import javax.jdo.Query

import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.servlet.http.HttpSession
import javax.servlet.ServletContext

/**
 * Wires up MetaClass methods in artifacts and servlet classes. Lots of code
 * borrowed from Grails.
 *
 * @author <a href='mailto:burt@burtbeckwith.com'>Burt Beckwith</a>
 */
class DynamicUtils {

	private static ServletContext _servletContext

	private static PersistenceManagerFactory _persistenceManagerFactory

	private static final ThreadLocal<PersistenceManager> _currentPersistenceManager =
		new ThreadLocal<PersistenceManager>()

	static PersistenceManagerFactory getPersistenceManagerFactory() {
		return _persistenceManagerFactory
	}

	static void configure(ServletContext servletContext) {
		_servletContext = servletContext
		_persistenceManagerFactory = JDOHelper.getPersistenceManagerFactory('transactions-optional')

		registerServletMetaclassMethods()

		String artifacts = new File('WEB-INF/artifacts.config').text
		artifacts.eachLine { line ->
			String[] parts = line.split('\\|')
			Class c = Class.forName(parts[1])
			switch (parts[0]) {
				case 'controllers':
					registerCommonWebProperties c, c.metaClass
					break
				case 'domain':
					configureDomainClass c, c.metaClass
					break
				case 'services':
					configureService c, c.metaClass
					break
			}
		}
	}

	private static void registerServletMetaclassMethods() {

		def getAttributeClosure = { String name ->
			def mp = delegate.class.metaClass.getMetaProperty(name)
			def result = null
			if (mp) {
				result = mp.getProperty(delegate)
			}
			else {
				result = delegate.getAttribute(name)
			}
			return result
		}

		def setAttributeClosure = { String name, value ->
			def mp = delegate.class.metaClass.getMetaProperty(name)
			if (mp) {
				mp.setProperty(delegate, value)
			}
			else {
				delegate.setAttribute(name, value)
			}
		}

		def getAttributeSubscript = { String key -> delegate.getAttribute(key) }

		def setAttributeSubScript = { String key, Object val -> delegate.setAttribute(key, val) }

		// enables access to servlet context with servletContext.foo syntax
		ServletContext.metaClass.getProperty = getAttributeClosure
		ServletContext.metaClass.setProperty = setAttributeClosure
		ServletContext.metaClass.getAt = getAttributeSubscript
		ServletContext.metaClass.putAt = setAttributeSubScript

		// enables access to session attributes using session.foo syntax
		HttpSession.metaClass.getProperty = getAttributeClosure
		HttpSession.metaClass.setProperty = setAttributeClosure
		// enables access to session attributes with session["foo"] syntax
		HttpSession.metaClass.getAt =  getAttributeSubscript
		// enables setting of session attributes with session["foo"] = "bar" syntax
		HttpSession.metaClass.putAt =  setAttributeSubScript

		HttpServletRequest.metaClass.isXhr = {->
			'XMLHttpRequest' == delegate.getHeader('X-Requested-With')
		}

		// enables access to request attributes with request["foo"] syntax
		HttpServletRequest.metaClass.getAt = { String key ->
			delegate.getAttribute(key)
		}
		// enables setting of request attributes with request["foo"] = "bar" syntax
		HttpServletRequest.metaClass.putAt = { String key, Object val ->
			delegate.setAttribute(key, val)
		}
		// enables access to request attributes using property syntax
		HttpServletRequest.metaClass.getProperty = getAttributeClosure
		HttpServletRequest.metaClass.setProperty = setAttributeClosure

		HttpServletRequest.metaClass.isGet = { -> delegate.method == "GET" }
		HttpServletRequest.metaClass.isPost = {-> delegate.method == "POST" }

		// enables searching of request attributes with request.find { it.key == 'foo' }
		HttpServletRequest.metaClass.find = { Closure c ->
			def request = delegate
			def result = [:]
			for (name in request.attributeNames) {
				def match = false
				switch (c.parameterTypes.length) {
					case 0:
						match = c.call()
						break
					case 1:
						match = c.call(key:name, value:request.getAttribute(name))
						break
					default:
						match = c.call(name, request.getAttribute(name))
				}
				if (match) {
					result[name] = request.getAttribute(name)
					break
				}
			}
			return result
		}

		// enables searching of for a number of request attributes using request.findAll { it.key.startsWith('foo') }
		HttpServletRequest.metaClass.findAll = { Closure c ->
			def request = delegate
		   	def results = [:]
			for (name in request.attributeNames) {
				def match = false
				switch (c.parameterTypes.length) {
					case 0:
						match = c.call()
						break
					case 1:
						match = c.call(key:name, value:request.getAttribute(name))
						break
					default:
						match =  c.call(name, request.getAttribute(name))
				}
				if (match) {
					results[name] = request.getAttribute(name)
				}
			}
			return results
		}

		// enables iteration over request attributes with each method request.each { name, value -> }
		HttpServletRequest.metaClass.each = { Closure c ->
			def request = delegate
			for (name in request.attributeNames) {
				switch (c.parameterTypes.length) {
					case 0:
						c.call()
						break
					case 1:
						c.call(key:name, value:request.getAttribute(name))
						break
					default:
						c.call(name, request.getAttribute(name))
				}
			}
		}

		// allows the syntax response << "foo"
		HttpServletResponse.metaClass.leftShift = { Object o ->
			delegate.writer << o
		}
	}

	private static registerCommonWebProperties(Class clazz, MetaClass mc) {
		mc.getParams = { -> RequestHolder.get().getParameterMap() }
		mc.getSession = { -> RequestHolder.get().request.session }
		mc.getRequest = { -> RequestHolder.get().request }
		mc.getResponse = { -> RequestHolder.get().response }
		mc.getServletContext = { -> _servletContext }

		addLogger clazz, mc
	}

	private static void configureDomainClass(Class clazz, MetaClass mc) {

		mc.save = { -> getCurrentPersistenceManager().makePersistent(delegate) }

		mc.isJdoPersistent = { -> JDOHelper.isPersistent(delegate) }

		mc.isJdoDeleted = { -> JDOHelper.isDeleted(delegate) }

		mc.isJdoDetached = { -> JDOHelper.isDetached(delegate) }

		mc.isJdoDirty = { -> JDOHelper.isDirty(delegate) }

		mc.isJdoNew = { -> JDOHelper.isNew(delegate) }

		mc.isJdoTransactional = { -> JDOHelper.isTransactional(delegate) }

		mc.getJdoTransactionalObjectId = { -> JDOHelper.getTransactionalObjectId(delegate) }

		mc.getJdoVersion = { -> JDOHelper.getVersion(delegate) }

		mc.getJdoObjectId = { -> JDOHelper.getObjectId(delegate) }

		mc.jdoMakeDirty = { String fieldName -> JDOHelper.makeDirty(delegate, fieldName) }

		mc.static.get = { key ->
			try {
				return getCurrentPersistenceManager().getObjectById(clazz, key)
			}
			catch (JDOObjectNotFoundException e) {
				return null
			}
		}

		mc.static.delete = { key ->
			withTransaction false, { pm ->
				def instance = pm.getObjectById(clazz, key)
				if (instance) {
					pm.deletePersistent(instance)
				}
			}
		}

		mc.static.withTransaction = { Closure c ->
			withTransaction(false, c)
		}

		mc.static.withTransaction = { boolean detachAllOnCommit, Closure c ->
			withTransaction(detachAllOnCommit, c)
		}

		mc.static.withPersistenceManager = { Closure c ->
			c(getCurrentPersistenceManager())
		}

		mc.static.findAll = { ->
			return getCurrentPersistenceManager().newQuery(clazz).execute()
		}

		mc.static.findAll = { Map m ->
			Query query = getCurrentPersistenceManager().newQuery(clazz)
			if (m.orderBy) {
				query.setOrdering(m.orderBy)
			}
			return query.execute()
		}

		addLogger clazz, mc
	}

	private static void configureService(Class clazz, MetaClass mc) {
		addLogger clazz, mc
	}

	private static void addLogger(Class c, MetaClass mc) {
		mc.getLog = { -> LogFactory.getLog(c) }
	}

	static PersistenceManager getCurrentPersistenceManager() {
		PersistenceManager current = _currentPersistenceManager.get()
		if (!current) {
			current = _persistenceManagerFactory.persistenceManager
			_currentPersistenceManager.set(current)
		}
		return current
	}

	static def withTransaction(boolean detachAllOnCommit, Closure c) {
		def pm = getCurrentPersistenceManager()
		def result
		try {
			pm.setDetachAllOnCommit(detachAllOnCommit)
			pm.currentTransaction().begin()
			result = c(pm)
			pm.currentTransaction().commit()
			return result
		}
		finally {
			try {
				if (pm.currentTransaction().isActive()) {
					pm.currentTransaction().rollback()
				}
				pm.close()
			}
			finally {
				_currentPersistenceManager.set(null)
			}
		}
	}
}
